"""

Simulations based on the Axelrod model and extensions to inspect the volatility of the features over time (AXELROD MODEL + Agreement threshold + two model variations based on the Social identity approach)
The Axelrod model is used to predict the number of changes per feature in comparison to the datasets and is used to
compare different model variations and their performace.

Input: Real data

--take the number of change per step and not the distance fucntion between the feature vectors as a break criteria
vector-version of axelrod --

Description:
The Axelrod model depicts convergence and diversity on a macro level, driven by local agent-based
interaction mechanisms. The agreement-threshold model (MacCarron et al. 2020a), an extension of the Axel-
rod model, acts as a multi-dimensional opinion dynamics model. We extend these agent-based
models by explicit aspects of the social identity approach to recover real-world dynamics better and to assess
the prediction performance of data simulations. We newly introduce mechanisms on in- and out-group interaction. 

Model variations:
Parameter: Group-dependent preference
    - Interaction within a group takes place without the limitation of the agreement threshold
    - Interaction between groups involves an agreement threshold
    - Group parameter needed

Parameter: In-group preference
    - Integrates an interaction preference towards in-group members
    - Reduced interaction probability with an out-group members
    - Inter-group interaction still possible, but unlikely
    - Group parameter needed


author: alejandro dinkelberg
date: 12.01.2022
"""
# -*- coding: utf-8 -*-
#   
# Variation of memory, agreement threshold, 4 different topologies

########################
# Import Libraries
#import matplotlib
#matplotlib.use('Agg')
import networkx as nx
import matplotlib.pyplot as plt
import random as rd
import pylab
from networkx.drawing.nx_agraph import graphviz_layout
import matplotlib.animation as animation
import numpy as np
from statistics import mean
from copy import deepcopy
from collections import Counter
import pandas as pd
from math import sqrt
import matplotlib

matplotlib.use('Agg')
from multiprocessing import Process, Queue, Pool
import datetime
from statistics import mean
import os



class AxelrodModelNumpy:

    def __init__(self, Grid_x=10, Grid_y=10, n_features=8, t_max=100, objective_changes_list=[1000] ,topo=0, at=0, m_AB=2, memory=False,
                 timepoint_to_measure=10000, data_array=None, group_array=None, only_same_group_interaction=False, group_dependent_at=False):
        
        # set grid size
        self.Grid_x = Grid_x
        self.Grid_y = Grid_y
        self.number_of_nodes = Grid_x * Grid_y

        # I/O for the following model variations:
        self.only_same_group_interaction = only_same_group_interaction  #Description: 10% probability that the agent can interact with a non-group member
        self.group_dependent_at = group_dependent_at                    #Description: self.at >= if not same group, otherwise at=0 <-- normal at
        
        # to collect information about cultural vector in system
        self.count_culture = dict()

        # number of features
        self.n_features = n_features
        
        # Choose total number of events
        self.t_max = t_max
        self.objective_changes = sum(objective_changes_list)
        self.objective_changes_list = objective_changes_list
        
        # topology
        self.topo =topo

        # edge thickness
        self.thickness = 5

        # define at which TIMEPOINTS to measure cluster size, number and smax -- Standard 10,000
        self.timepoint_to_measure = timepoint_to_measure

        # Agreement threshold
        self.at = at
        self.at_initial = at
        
        # determine who belongs to which group
        self.group_array = group_array
        
        # number of changes
        self.changes = 0
        self.changes_feature = {f: 0 for f in range(n_features)}

        # Boolian, if t-1 status is integrated
        self.memory = memory

        # return variable
        self.timepoints = {}

        # This is how much to incriment the x for each trait box (reduce for lots of traits)
        self.xi = 1.2

        # line_style -- need to make this longer than n_features
        self.line_style = ['-.', ':', '--', '-', '-']

        #########################

        # turns interactive on
        # pylab.ion()
        self.fig, (ax1, ax2) = plt.subplots(1, 2, figsize=(18, 9), facecolor='0.15')
        ax1.set_axis_off()
        ax2.set_axis_off()

        # draw initial nodes with diff topology
        if topo == 1:
            self.G = nx.barabasi_albert_graph(Grid_x*Grid_y, m_AB)
        elif topo == 2:
            self.G = nx.complete_graph(Grid_y*Grid_x)
        elif topo == 3:
            self.G = nx.watts_strogatz_graph(n=Grid_x*Grid_y, k=4, p=0.5)
        elif topo == 4:
            self.G = nx.gnm_random_graph(Grid_y*Grid_x, (2*(Grid_x*Grid_y)-Grid_x-Grid_y))  # m= edges in raster y * x
        else:
            self.G = nx.grid_2d_graph(Grid_x, Grid_y)

        # convert the network structure to array/ could also use array from nx
        G_integernodes = nx.convert_node_labels_to_integers(self.G)
        self.nodes_array = np.array([np.array([z for z in G_integernodes.neighbors(i)]) for i in G_integernodes.nodes()], dtype=object) #dtype=object to avaoid deprecaton warning
        
        #############
        # initial edges WITHOUT DATA

        #traits = range(n_traits)
        #trait_min, trait_max = min(traits), max(traits)

        #self.features = np.array([[rd.randint(trait_min, trait_max) for _ in range(n_features)] for u in range(self.number_of_nodes)])
        #print(type(self.features), self.features)
        
        # ------------------------ CSV seed -------------------------------------

        # seed with real Feature_traits
        self.data_array = data_array
        self.data_list = self.data_array

        # set up and define data
        
        self.features = {}
        self.features_arrays = []
        
        #randomize selection of the features
        possible_features = list(range(len(self.data_list)))
                                 
        # feed axelrod with data
        self.group_affiliations = []
        for u in self.G.nodes():
            # which element of the list should be selected?
            i = possible_features.pop(rd.randrange(len(possible_features)))
            
            # collect data for i
            self.features[u] = self.data_list[i]
            self.group_affiliations.append(self.group_array[i])    ## reorder the group affiliation

            self.features_arrays.append(np.array(self.data_list[i]))
        
        self.group_affiliations = np.array(self.group_affiliations)    ## reorder the group affiliation

        # change to array (depricated)
        #self.group_affiliations = np.array(self.group_affiliations)
        
        self.features_arrays = np.array(self.features_arrays, dtype=object) # final array of features; has to have the numpy structure
        # ------------------------ CSV seed end----------------------------------
        
        # set features
        self.features = self.features_arrays
        
        # Copy features FIRST
        self.features_copy = deepcopy(self.features)
        self.features_initial = deepcopy(self.features)

        # RUN the model
        event = self.run_model(t_max)

    def run_model(self, t_max):
        """
        Run method of the Axelrod model, is dependent on the number of changes from the data and will stop when the maximum number
        of possible changes is reached. We only set a limit to secure that it terminates but usually it should not be reached. 
        The model runs and gives the results for 
        :param t_max: number of events
        :return: set of measurements --> list(number of changes per feature, length of run, number of overall changes, biggest cluster size)
        """
        
        # save at the right moment
        activator = 0
        save_step = False
        for event in range(1, t_max + 1):

            # Compute and save the distribution of starting vectors (distribution for each feature)
            if event == 1:            # data input
                
                attitude_distribution = []
                for i in range(self.n_features):
                    attitude_distribution.append([0] * np.max(self.features[:,i]))  # highest value of traits is also number of traits 0-4 or 0-8
                attitude_distribution_zeros = deepcopy(attitude_distribution)

      
                for node in self.features:
                    for i, trait in enumerate(node):
                        attitude_distribution[i][trait-1] += 1      # attention: for realdata [trait-1] !


            # pick a random node and a random neighbour
            u = rd.randint(0, self.number_of_nodes-1)
            u_neigbours = self.nodes_array[u]
            if self.topo == 4:
                while len(u_neigbours) == 0:
                    u = rd.randint(0, self.number_of_nodes - 1)
                    u_neigbours = self.nodes_array[u]
            v = np.random.choice(u_neigbours)

            # model variant activated? # Only interact if they are in the same group
            if self.only_same_group_interaction:
                # interaction probability if they do not belong to the same group, 10%
                if self.group_affiliations[u] != self.group_affiliations[v]:
                    if rd.random() > 0.1:  # value 0.1 == 10% could be changed to lower/higher value
                        continue
            
            # get the number of common features and the index of non-common features
            f_u, f_v = self.features[u], self.features[v]

            # interaction?
            if not np.array_equal(f_u, f_v):    # do not interact if already identical features
                
                # WITH agreement threshold
                if self.at != 0:
                    # Agreement Threshold is ON
                    # --------------         --------------------
                    # Get the difference between the two vectors,
                    # then figure out what they have in common
                    # that is withing the agreement threshold
                    diff = np.absolute(f_u - f_v)
                    common = diff[diff <= self.at].size / self.n_features
                    within = np.where((diff <= self.at) & (diff > 0))[0]

                    # at is defined by the group parameter; more variability allowed when the group is the own group;  if not then use the agreement threshold
                    if self.group_dependent_at and self.group_affiliations[u] == self.group_affiliations[v]:
                        # if group_dependent_at is 
                        # only is the agents belong to the same group it will be different
                        self.at = 0
                        # Agreement threshold is OFF
                        # Now I have to runt the AT=0 version
                        non_common_features_index = np.array(*np.where(f_u != f_v))  # was an 5x1 array
                        # with probability of common features change one of the non-common features of u
                        if (prob := rd.random()) <= common:
                            feature_to_change = np.random.choice(non_common_features_index)
                            self.features[(u, feature_to_change)] = self.features[(v,feature_to_change)]
                            
                            # count changes
                            self.changes_feature[feature_to_change] += 1
                            self.changes += 1
                    else:
                        # NORMAL version
                        # with probability of common features change one of the non-common features of u
                        prob = rd.random()
                        if (prob := rd.random()) <= common and within.size > 0:
                            
                            feature_to_change = rd.choice(within)
                            self.features[(u, feature_to_change)] = self.features[(v,feature_to_change)]
                            # count changes
                            self.changes_feature[feature_to_change] += 1
                            self.changes += 1
                else:
                    # Agreement threshold is OFF
                    non_common_features_index = np.array(*np.where(f_u != f_v))  # was an 5x1 array

                    # non_common_features = [i for i, j in zip(f_u, f_v) if i != j]
                    common = (self.n_features - len(non_common_features_index)) / self.n_features

                    # with probability of common features change one of the non-common features of u
                    if (prob := rd.random()) <= common:
                        feature_to_change = np.random.choice(non_common_features_index)
                        self.features[(u, feature_to_change)] = self.features[(v,feature_to_change)]
                        
                        # count changes
                        self.changes_feature[feature_to_change] += 1
                        self.changes += 1
                        
            # measure of cluster size, number and smax
            self.distance_from_original = np.sum(np.abs(self.features_initial - self.features))
            #print(self.distance_from_original, self.objective_changes_list[-1])
            #activator
            if activator < len(self.objective_changes_list)-1 and self.objective_changes_list[activator] <= self.changes < self.objective_changes_list[activator+1]:
                #print('##########', self.distance_from_original, event)
                save_step = True
                activator += 1 
            
            
            # save certain at certain steps and stop if number of changes is reached
            if save_step or self.changes >= self.objective_changes_list[-1]:  #self.distance_from_original == self.objective_changes or                                                   #event % self.timepoint_to_measure == 0 or event == t_max:
                
                # Copy Graph every  timepoints_to_measure=10000 steps
                self.features_copy = deepcopy(self.features)

                # generate features list
                feature_list = [' '.join(str(i) for i in feat) for feat in list(self.features)]
 
                # This counts the number of times each feature string appears
                clusters = [feature_list.count(s) for s in set(feature_list)]
                                        # mean_cluster size     &  num of cluster & S_max
                
                
                # calculate attitude distribution
                attitude_distribution = deepcopy(attitude_distribution_zeros)
                for node in self.features:
                    for i, trait in enumerate(node):
                        attitude_distribution[i][trait-1] += 1 #for realdata [trait-1] !
                
                self.count_culture[self.distance_from_original] = attitude_distribution
                
                # return list of results
                self.timepoints[self.distance_from_original] = [round(mean(clusters), 2),
                                                                len(clusters),
                                                                max(clusters),
                                                                self.changes,
                                                                event,
                                                                deepcopy(self.changes_feature),
                                                                attitude_distribution]
                # end here
                if self.changes >= self.objective_changes_list[-1]: # if done
                    return event
                
                # reset save step
                save_step = False
               # print(self.timepoints.values())
        #print(self.features)
        return event


def number_of_changes(dataset, precise_difference=True):
    """ define number of changes and give back a dictionary

    Args:
        dataset (string): filename
        precise_difference (bool, optional): precise difference (forth and back are two changes) Defaults to True.

    Returns:
        _type_: (dict) differences, (dict) used data, ID_party_aff, diff_array
    """
    csv_file = dataset[0] # string
    n_items_per_timepoint = dataset[1] # int 
    timepoints = dataset[2] # int
    
    
    df = pd.read_csv(csv_file)
    
    # get ID and party aff 
    ID_party_aff = df[['ID'] + list(df.filter(regex = 'party_aff*')) + list(df.filter(regex = 'Comp*'))]
    print(ID_party_aff)
    # drop it for the rest of the analysis
    df.drop(labels=['Unnamed: 0','ID'], axis=1, inplace=True)         # not relevant here
    df.drop(list(df.filter(regex = 'party_aff*')) + list(df.filter(regex = 'Comp*')), axis = 1, inplace = True)    # drop party_aff

    
    data_dict = {}
    for t in range(1,timepoints+1):
        data_dict[t] = np.array(df.values.tolist())[:, (t-1)*n_items_per_timepoint:t*n_items_per_timepoint]

    # check if all arrays same size:
    check_size = list(data_dict.values())[0].shape
    for value in data_dict.values():
        if check_size != value.shape:
            print('/= 42!')
    diff_dict = {}
    diff_array = {}
    for t in range(1,timepoints):
        diff_array[t] = np.absolute(data_dict[t] - data_dict[t+1]) if precise_difference else np.array(np.absolute(data_dict[t] - data_dict[t+1]), dtype=bool)
        diff_dict[f'T{t+1}'] = sum(sum(diff_array[t]))

    return diff_dict, data_dict, ID_party_aff, diff_array


def feature_distribution(n_features=8, features=None, group_array=None):
    """get feature distribution for each group

    Args:
        n_features (int, optional): _description_. Defaults to 8.
        features (_type_, optional): _description_. Defaults to None.
        group_array (_type_, optional): _description_. Defaults to None.

    Raises:
        ValueError: ("group and features do not have the same length")

    Returns:
        _type_: (dict) with attitude distributions
    """
    
    attitude_distribution = []
    for i in range(n_features):
        attitude_distribution.append([0] * np.max(features[:,i]))  # highest value of traits is also number of traits 0-4 or 0-8
    
    group_attitude_distribution = {}
    for group in list(set(group_array)):
        group_attitude_distribution[group] = deepcopy(attitude_distribution)
    
    if len(group_array) != len(features):
        raise ValueError("group and features do not have the same length")
    
    for group, node in zip(group_array, features):
        for i, trait in enumerate(node):
            group_attitude_distribution[group][i][trait-1] += 1 # for realdata [trait-1] !!!!!!!!!!!!!

    return group_attitude_distribution
    

def run_model(grid, n_features, t_max, objective_changes_list, topo, at, connection, timepoint_to_measure, data_array, group_array, only_same_group_interaction, group_dependent_at, q):
    """ callable run function for multiprocessing

    Args:
        grid (int): size of grid
        n_features (int): number of features
        t_max (int): max running time
        objective_changes_list (list): maximum number of changes which are simulated by the mode
        topo (int): underlying topology of the model
        at (int): agreement threshold; 0 = no agreement threshold
        connection (int): number of links between agents (for example to generate AB network)
        timepoint_to_measure (int): How often do we measure the state of the model
        data_array (np.array): position of each agent from the data
        group_array (np.array): group identifies for every agent
        only_same_group_interaction (bool): model variation; interaction highly limited to group members
        group_dependent_at (bool): model variation; at for out-group members and no at for in-group members
        q (queue): multiprocessing, get back the results (return statement)
    """
    
    amodel = AxelrodModelNumpy(Grid_x=grid, Grid_y=grid, n_features=n_features,t_max=t_max, objective_changes_list=objective_changes_list,  
                              topo=topo,
                              at=at, m_AB=connection, timepoint_to_measure=timepoint_to_measure, data_array=data_array,  group_array=group_array,
                              only_same_group_interaction=only_same_group_interaction,
                              group_dependent_at=group_dependent_at)
    
    # collect results
    a = []       
    for distance_from_original, timepoint in amodel.timepoints.items():
        
        a.append(timepoint[0])
        a.append(timepoint[1])
        a.append(timepoint[2])
        a.append(distance_from_original)#(v[0])
        a.append(timepoint[3])#(v[1])
        a.append(timepoint[4])#(v[2]) == counter for the event, when is the amount of changes reached
        a.append(timepoint[5])#(i)
        a.append(timepoint[6])
    a.append(feature_distribution(amodel.n_features, amodel.features, amodel.group_affiliations)) # group distribution of the features; divided 
    
    #return results
    q.put(a)

def count_cluster(f):
    """ count the number of clusters (agents with identical vectors)
        #no used#
    Args:
        f (np.array): cultural vectors of the agents

    Returns:
        list: mean cluster size, number of clusters
    """
    feature_list = [''.join(str(i) for i in feat) for feat in list(f)]
    # This counts the number of times each feature string appears
    clusters = [feature_list.count(s) for s in set(feature_list)]
    return [round(mean(clusters), 2), len(clusters)]


def run_main(selected_dataset = 'BJSP', at_value = [1], topo_value = [0], rounds=2000, only_same_group_interaction=False, group_dependent_at=False):
    """ main function

    Args:
        selected_dataset (str, optional): determine dataset. Defaults to 'BJSP'.
        at_value (list, optional): agreement threshold. Defaults to [1].
        topo_value (list, optional): topology of the model. Defaults to [0].
        rounds (int, optional): max number of simulation runs. Defaults to 2000.
        only_same_group_interaction (bool, optional): model variation 1. Defaults to False.
        group_dependent_at (bool, optional): model variation 2. Defaults to False.
    """
    
    # which data set?
    datasets = {'ANES': ['ANES_2016_2020_same_participants.csv', 8, 2, 'party_aff_2016'],
                'ANES_only_DemRep': ['ANES_2016_2020_same_participants_only_dem_and_rep.csv', 8, 2, 'party_aff_2016'],      # around 100 participants
               'BJSP': ['Anon_T1_to_T3_same_participants.csv', 8, 3, 'Comp'],
               'BJSP_vax': ['Anon_T1_to_T3_with_vax_same_participants.csv', 11, 3, 'Comp'], 
               'BJSP_T1_to_T3_with_vax_all4pointscale_same_participants': ['Anon_T1_to_T3_with_vax_all4pointscale_same_participants.csv', 11, 3, 'Comp'],   # same number of traits for all
               'USelection_improved_group_at': ['Paul_T1-4_same_participants.csv', 8, 4, 'party_aff_T2']}
    
    # Count the number of changes in the data set
    diff_dict, data_dict, ID_party_aff, diff_array = number_of_changes(dataset=datasets[selected_dataset], precise_difference=False)
        # precise_difference=False has to be used!
        # in Axelrod I also have to count the changes made and not the difference, like this it gets more consistent
        # if Axelrod starts from T1 and measures the changes always from T1 the diff_dict has be calculated differently 
        # not decided on that one

    # get the changes
    dict_changes = {}
    for t, array in diff_array.items():
        dict_changes[t] = {}
        for i in range(len(array[0])):
            dict_changes[t][i] = np.sum(array[:,i]) 
    pd.DataFrame(dict_changes).to_csv(f'{selected_dataset}_changes_timepoints.csv')

        
    # size  of grid
    grid = int(sqrt(len(data_dict[1]))) 
    
    # distance from initial features
    objective_changes_list = []                                         # must be defined depending on steps
    temp = 0
    for i in diff_dict.values():
        temp += i #+ 100000
        objective_changes_list.append(temp) # list sums up the changes of all previous ones
    
    print('# Number of changes in the data: ', objective_changes_list)

    # to save results and the processes
    processes = []



    # ----------defining vars
    
    # connections
    connection = 2
    
    # timepoints to measure
    timepoint_to_measure = 200000        # high enough to reduce calculations

    #options:
    add_f_name = ''

    # add model variat to filename
    if only_same_group_interaction:
        add_f_name += 'only_same_group_interaction'

    # add model variat to filename
    if group_dependent_at:
        add_f_name += 'new_grp_dep_AT'

    # boolean list to integrate in for-loop
    memory_value = [False]

    # agreement threshold --> 0 = no agreement threshold
    # topology 0=std, 1= AB, 2=complete, 3=small-world, 4=random

    # get feature distribution from the data input
    T1_group_attitude_distribution = feature_distribution(n_features=datasets[selected_dataset][1], features=data_dict[1], group_array=ID_party_aff[datasets[selected_dataset][3]].array)
    T2_group_attitude_distribution = feature_distribution(n_features=datasets[selected_dataset][1], features=data_dict[2], group_array=ID_party_aff[datasets[selected_dataset][3]].array)
    print(T2_group_attitude_distribution)
    df = pd.DataFrame([str(T1_group_attitude_distribution)]).to_csv(f'{selected_dataset}_T1_feature_dist.csv')
    df = pd.DataFrame([str(T2_group_attitude_distribution)]).to_csv(f'{selected_dataset}_T2_feature_dist.csv')
    
    
    error_count = 0
    # for topology// in this variation normally only len of 1
    for topo in topo_value:
        # at ==> 0 or 1
        for at in at_value:
            new_l = list()
            # number of runs =  i * number of processes
            
            # 'where to save
            if topo == 1:
                topology = "barabasi_albert_graph"
            elif topo == 2:
                topology = "complete_graph"
            elif topo == 3:
                topology = "watts_strogatz"
            elif topo == 4:
                topology = "ER-random"      
            else:
                topology = "std"

            # connection
            m_AB = "" if connection == 2 else "_m_AB_" + str(connection)

            columns = ['Cluster size', 'Number of clusters', 'S_max', 'distance_from_original','total_changes','time','changes_features','dist_features'] * len(objective_changes_list) + ['group_dist_features']
            df = pd.DataFrame(new_l, columns=columns)
            
            # timestamp
            t = str(datetime.datetime.now().ctime())
            t = t.replace(":", "_")
            t = t.replace(' ', "_")

            temp = f"{name}/{selected_dataset}_N_" + str(grid * grid) + topology + m_AB + "_axlrd_t_" + "_at=" + f'{at}' + t + add_f_name + ".csv"
            #writer = pd.ExcelWriter(temp, engine='xlsxwriter')  # Werden im Pfad gespeichert.
            
            # number of rounds
            for i in range(rounds):
                # number of processes
                # max time to evolve     # num of events / time
                t_max = 20000000 #rd.randint(int(diff_dict['T2']*0.8), int(diff_dict['T2']*1.2))    # plus minus 20%

                # collect results from multiprocessing
                x = Queue()
                
                # number of processes per round
                for _ in range(5):  # not more than 5; Queue receives to much info
                    p = Process(target=run_model, args=(grid, datasets[selected_dataset][1],t_max, objective_changes_list, topo, at, connection, timepoint_to_measure, data_dict[1],
                                                        ID_party_aff[datasets[selected_dataset][3]].array, only_same_group_interaction, group_dependent_at, 
                                                        x))
                    p.start()
                    processes.append(p)
                    # print(p.pid)
                    # new_l.append(x.get()) # not working!
                for p in processes:
                    p.join()
                while not x.empty():
                    new_l.append(x.get())

                # save in filter
                try:
                    df = pd.DataFrame(new_l, columns=columns)
                    # to don't loose data
                    if (i %10 == 0 and i>1) or (i == rounds-1):
                        print(temp, ': ', (i/rounds)*100,'%')
                        df.to_csv(temp)#, 'Sheet1')
                
                except ValueError:
                    error_count +=1
                    print('probably only one column in the data, reached the end?', error_count)

            # final save
            #df.to_csv(temp) #, 'Sheet1')
            #writer.save()
            print(temp, f"saved at {datetime.datetime.now().ctime()}")
            

if __name__ == '__main__':
    
    # run through datasets  and simulate
    for name in ['ANES_only_DemRep']:   #'BJSP_T1_to_T3_with_vax_all4pointscale_same_participants']:#['USelection_improved_group_at']: #
        
        if not os.path.isdir(f'{name}'):
            os.mkdir(f'{name}') 
        print('run ', name)
        for topo in [0]: #,1,2,3,4] 
            rounds=2000
            run_main(selected_dataset = name, at_value = [0, 1], topo_value = [topo], rounds=rounds, only_same_group_interaction=False, group_dependent_at=False)
            run_main(selected_dataset = name, at_value = [1], topo_value = [topo], rounds=rounds, only_same_group_interaction=False, group_dependent_at=True)
            run_main(selected_dataset = name, at_value = [0,1], topo_value = [topo], rounds=rounds, only_same_group_interaction=True, group_dependent_at=False)   